"""numarray: The big enchilada numeric module

$Id: numarray.py,v 1.36 2002/06/17 14:00:20 jaytmiller Exp $
"""
import sys as _sys
import types, math, os.path
import memory
import ndarray
import _bytes
import _numarray
import _sort
import ufunc
import operator as _operator
import copy as _copy

from ufunc import *
import arrayprint
from numerictypes import *
from ndarray import *
from math import pi, e

# rename built-in function type so not to conflict with keyword
_type = type

MAX_LINE_WIDTH = None
PRECISION = None
SUPPRESS_SMALL = None

_FPE_DIVIDE_BY_ZERO = 1
_FPE_OVERFLOW       = 2
_FPE_UNDERFLOW      = 4
_FPE_INVALID        = 8

class NumError:
    """Defines how numeric errors should be handled."""
    def __init__(self, all="warn",
          overflow=None, underflow=None, dividebyzero=None, invalid=None):
        self._errorset = ("ignore", "warn", "raise")
        self.setMode(all=all, underflow=underflow, overflow=overflow,
                     dividebyzero=dividebyzero, invalid=invalid)

    def setMode(self, all="warn", overflow=None, underflow=None,
                      dividebyzero=None, invalid=None):
        if all in self._errorset:
            self.overflowMode = all
            self.underflowMode = all
            self.dividebyzeroMode = all
            self.invalidMode = all
        if overflow in self._errorset:
            self.overflowMode = overflow
        if underflow in self._errorset:
            self.underflowMode = underflow
        if dividebyzero in self._errorset:
            self.dividebyzeroMode = dividebyzero
        if invalid in self._errorset:
            self.invalidMode = invalid

Error = NumError()

class MathDomainError(ArithmeticError): pass
class UnderflowError(ArithmeticError): pass

def handleError(errorStatus, sourcemsg):
    """Take error status and use error mode to handle it."""
    if errorStatus & _FPE_INVALID:
        if Error.invalidMode == "warn":
            print "Warning: Encountered invalid numarray result(s)", sourcemsg
        if Error.invalidMode == "raise":
            raise MathDomainError(sourcemsg)
    if errorStatus & _FPE_DIVIDE_BY_ZERO:
        if Error.dividebyzeroMode == "warn":
            print "Warning: Encountered divide by zero(s)", sourcemsg
        if Error.dividebyzeroMode == "raise":
            raise ZeroDivisionError(sourcemsg)
    if errorStatus & _FPE_OVERFLOW:
        if Error.overflowMode == "warn":
            print "Warning: Encountered overflow(s)", sourcemsg
        if Error.overflowMode == "raise":
            raise OverflowError(sourcemsg)
    if errorStatus & _FPE_UNDERFLOW:
        if Error.underflowMode == "warn":
            print "Warning: Encountered underflow(s)", sourcemsg
        if Error.underflowMode == "raise":
            raise UnderflowError(sourcemsg)

PyINT_TYPES = {
            types.IntType: 1,
            types.LongType: 1,
            }
PyREAL_TYPES = {
            types.IntType: 1,
            types.LongType: 1,
            types.FloatType: 1,
            }

# Python numeric types with values indicating level in type hierarchy
PyNUMERIC_TYPES = {
            types.IntType: 0,
            types.LongType: 1,
            types.FloatType: 2,
            types.ComplexType: 3
            }

# Mapping back from level to type
PyLevel2Type = {}
for key, value in PyNUMERIC_TYPES.items():
    PyLevel2Type[value] = key
del key, value

# Mapping from Python to Numeric types
Py2NumType = {
            types.IntType:   Long,
            types.LongType:  Long,
            types.FloatType: Float,
            types.ComplexType: Complex
            }

def array2list(arr):
    #XXX Probably should make this a tolist method of NumArray (and write in C)
    retlist = []
    if len(arr._shape) <= 1:
        for item in arr:
            retlist.append(item)
    else:
        for subarray in arr:
            retlist.append(array2list(subarray))
    return retlist

# array factory functions

def _flattenseq(seq):
    """Turn a nested set of sequences into a list of non sequence values.

    Returns list [seq] if argument is not a sequence.
    """
    #XXX Should be written in C for efficiency
    #XXX Use Tkinter._flatten as model -- it is vastly faster
    try:
        stypes = map(type, seq)
    except TypeError:
        return [seq]
    if types.ListType not in stypes and types.TupleType not in stypes:
        return seq
    retlist = []
    tt = (types.TupleType, types.ListType)
    for i in xrange(len(seq)):
        if stypes[i] in tt:
            retlist.extend(_flattenseq(seq[i]))
        else:
            retlist.append(seq[i])
    return retlist

def _maxtype(args):
    """Find the maximum scalar numeric type in the arguments.

    An exception is raised if the types are not python numeric types.
    """
    if not len(args):
        return None
    tlevel = map(PyNUMERIC_TYPES.get, map(type, args))
    if None in tlevel:
        i = tlevel.index(None)
        raise TypeError("Expecting a python numeric type, got a %s" %
            _type(args[i]).__name__)
    return PyLevel2Type[max(tlevel)]

def _storePyValueInBuffer(buffer, Ctype, index, value):
    """Store a python value in a buffer, index is in element units, not bytes"""
    # Do not use for complex scalars!
    Ctype._conv.fromPyValue(value, buffer._data,
                            index*Ctype.bytes, Ctype.bytes, 0)

def _storePyValueListInBuffer(buffer, Ctype, valuelist):
    # Do not use for complex values!
    for i in xrange(len(valuelist)):
        _storePyValueInBuffer(buffer, Ctype, i, valuelist[i])

def _fillarray(size, start, delta, type=None):
    ptype = _maxtype((start, delta))
    if PyINT_TYPES.has_key(ptype):
        ptype = Long
    elif PyREAL_TYPES.has_key(ptype):
        ptype = Float
    else:
        ptype = Complex
    if type:
        outtype = _getType(type)
        if _isComplexType(ptype) and not _isComplexType(outtype):
            raise TypeError("outtype must be a complex type")
    else:
        outtype = ptype
    if _isComplexType(outtype):
        # Not memory efficient at the moment
        real = _fillarray(size, complex(start).real, complex(delta).real,
                type = _realtype(ptype))
        imag = _fillarray(size, complex(start).imag, complex(delta).imag,
                type = _realtype(ptype))
        outarr = ComplexArray((size,), outtype)
        outarr.getreal()[:] = real
        outarr.getimag()[:] = imag
    else:
        # save parameters in a buffer
        parbuffer = ufunc._bufferPool.getBuffer()
        _storePyValueListInBuffer(parbuffer, ptype, [start, delta])
        cfunction = _sort.functionDict[repr((ptype.name, 'fillarray'))]
        outarr = NumArray((size,), outtype)
        if ptype == outtype:
            # no conversion necessary, simple case
            _numarray.CheckFPErrors()
            cfunction(size, 1, 1, ((outarr._data, 0), (parbuffer._data, 0)))
            errorstatus = _numarray.CheckFPErrors()
            if errorstatus:
                handleError(errorstatus, " in fillarray")
        else:
            # use buffer loop
            convbuffer = ufunc._bufferPool.getBuffer()
            convfunction = ptype._conv.astype[outtype.name]
            bsize = len(convbuffer._data)/ptype.bytes
            iters, lastbsize = divmod(size, bsize)
            _numarray.CheckFPErrors()

            outoff = 0
            for i in xrange(iters + (lastbsize>0)):
                if i == iters:
                    bsize = lastbsize
                cfunction(bsize, 1, 1,
                          ((convbuffer._data, 0), (parbuffer._data, 0)))
                convfunction(bsize, 1, 1, 
                             ((convbuffer._data, 0), (outarr._data, outoff)))
                outoff += bsize*outtype.bytes
                start += delta * bsize
                _storePyValueListInBuffer(parbuffer, ptype, [start, delta])
            errorstatus = _numarray.CheckFPErrors()
            if errorstatus:
                handleError(errorstatus, " in fillarray")
    return outarr

def NewArray(buffer=None, shape=None, type=None):
    if _isComplexType(type):
        return ComplexArray(buffer=buffer, shape=shape, type=type)
    else:
        return NumArray(buffer=buffer, shape=shape, type=type)

def _frontseqshape(seq):
    """Find the length of all the first elements, return as a list"""
    if not len(seq):
        return (0,)
    try:
        shape = []
        while 1:
            shape.append(len(seq))
            seq = seq[0]
    except TypeError:
        return shape

def _setarray(arr, seq, indices):
    """Populate array with sequence values starting at specified partial indices"""
    try:
        depth = len(indices)
        if len(arr._shape) == depth+1:
            # Must be at lowest level
            for i in xrange(arr._shape[depth]):
                arr[tuple(indices+[i])] = seq[i]
        else:
            for i in xrange(arr._shape[depth]):
                _setarray(arr, seq[i], indices+[i])
    except IndexError:
        raise IndexError("Supplied sequence not shape consistent.")

def fromlist(seq, type=None, shape=None):
    """fromlist(seq, type=None, shape=None) creates a NumArray from
    the sequence 'seq' which must be a list or tuple of python numeric
    types.  If type is specified, it is as the type of the resulting
    NumArray.  If shape is specified, it becomes the shape of the result
    and must have the same number of elements as seq.
    """
    nums = map(lambda x: isinstance(x, ndarray.NDArray), seq)
    if nums:
        if reduce(_operator.and_, nums):
            l = list(seq)
            for i in range(len(l)):
                l[i] = l[i].view()
                l[i].setshape((1,)+l[i].getshape())
            arr = ndarray.concatenate(l)
            if type is not None:
                arr = arr.astype(type)
            if shape is not None:
                arr.setshape(shape)
            return arr
        elif reduce(_operator.or_, nums):
            raise TypeError("Cannot mix arrays and non-arrays...")

    if not len(seq) and type is None:
        type = Long
        
    highest_type = _maxtype(_flattenseq(seq))
    tshape = _frontseqshape(seq)
    if shape is not None and ndarray.product(shape) != ndarray.product(tshape):
        raise ValueError("shape incompatible with sequence")
    ndim = len(tshape)
    if ndim <= 0:
        raise TypeError("Argument must be a sequence")
    if type is None:
        type = Py2NumType.get(highest_type)
    if type is None:
        raise TypeError("Cannot create array of type %s" % highest_type.__name__)
    tshape = tuple(tshape)
    arr = NewArray(shape=tshape, type=type)
    _setarray(arr, seq, [])
    if shape is not None:
        arr.setshape(shape)
    return arr

def array(buffer=None, type=None, shape=None,
          copy=1, savespace=0, typecode=None):
    """array(buffer=None, type=None, shape=None, copy=1) constructs a
    NumArray by calling NumArray, one of its factory functions
    (fromstring, fromfile, fromlist), or by making a copy.
    """
    if typecode and type and (typecode != type):
        raise ValueError("Cannot specify both typecode and type.")
    elif typecode:
        type = _getType(typecode)
    if isinstance(shape, types.IntType):
        shape = (shape,)
    if buffer is None and (shape is None or type is None):
        raise ValueError("Must define shape and type if buffer is None")
    elif buffer is None or ndarray.SuitableBuffer(buffer):
        return NewArray(buffer=buffer, shape=shape, type=type)
    elif isinstance(buffer, types.StringType):
        return fromstring(buffer, shape=shape, type=type)
    elif (isinstance(buffer, types.ListType) or
          isinstance(buffer, types.TupleType)):
        return fromlist(buffer, type, shape)
    elif PyNUMERIC_TYPES.has_key(_type(buffer)):
        if shape and shape != ():
            raise ValueError("Only shape () is valid for a rank 0 array.")
        return fromlist([buffer], shape=(), type=type or
                        Py2NumType[_type(buffer)])
    elif isinstance(buffer, NDArray):
        if (type and (buffer.type() != type)) or copy:
            a = buffer.astype(type) # make a re-typed copy
        else:
            a = buffer
        if shape is not None:
            a.setshape(shape)
        return a
    elif isinstance(buffer, types.FileType):
        return fromfile(buffer, type=type, shape=shape)
    else:
        raise ValueError("Unknown input type")

def inputarray(seq, type=None, typecode=None):
    """inputarray(seq, type=None)  converts scalars, lists and tuples to
    arrays if possible, passes NumArrays thru making copies only to convert
    types, and raises a TypeError for everything else.
    """
    return array(seq, type=type, typecode=typecode, copy=0)

asarray = inputarray   # Alias for Numeric compatability

def _getType(type):
    """Return the numeric type object for type

    type may be the name of a type object or the actual object
    """
    if isinstance(type, NumericType):
        return type
    try:
        return typeDict[type]
    except KeyError:
        raise TypeError("Not a numeric type")

def fromstring(datastring, type, shape=None):
    """Create an array from binary data contained in a string (by copying)"""
    type = _getType(type)
    if not shape:
        size, rem = divmod(len(datastring), type.bytes)
        if rem:
            raise ValueError("Type size inconsistent with string length")
        else:
            shape = (size,) # default to a 1-d array
    elif _type(shape) is types.IntType:
        shape = (shape,)
    if len(datastring) != (ndarray.product(shape)*type.bytes):
        raise ValueError("Specified shape and type inconsistent with string length")
    arr = NewArray(shape=shape, type=type)
    strbuff = buffer(datastring)
    nelements = arr.nelements()
    # Currently uses only the byte-by-byte copy, should be optimized to use
    # larger element copies when possible.
    cfunc = _bytes.functionDict["copyNbytes"]
    cfunc((nelements,), strbuff, 0, (type.bytes,),
          arr._data, 0, (type.bytes,), type.bytes)
    if arr._type == Bool:
        arr = ufunc.not_equal(arr, 0)
    return arr

def fromfile(file, type, shape=None):
    """Create an array from binary file data

    If file is a string then that file is opened, else it is assumed
    to be a file object. No options at the moment, all file positioning
    must be done prior to this function call with a file object

    """
    type = _getType(type)
    name =  0
    if _type(file) == _type(""):
        name = 1
        file = open(file, 'rb')
    size = os.path.getsize(file.name) - file.tell()
    if not shape:
        nelements, rem = divmod(size, type.bytes)
        if rem:
            raise ValueError(
                "Type size inconsistent with shape or remaining bytes in file")
        shape = (nelements,)
    elif _type(shape) is types.IntType:
        shape=(shape,)
    nbytes = ndarray.product(shape)*type.bytes
    if nbytes > size:
        raise ValueError(
                "Not enough bytes left in file for specified shape and type")
    # create the array
    arr = NewArray(shape=shape, type=type)
    # Use of undocumented file method! XXX
    nbytesread = memory.file_readinto(file, arr._data)
    if nbytesread != nbytes:
        raise IOError("Didn't read as many bytes as expected")
    if name:
        file.close()
    if arr._type == Bool:
        arr = ufunc.not_equal(arr, 0)
    return arr

class NumArray(NDArray):
    """Fundamental Numeric Array
    
    type       The type of each data element, e.g. Int32  
    byteorder  The actual ordering of bytes in buffer: "big" or "little".
    """

    def __init__(self, shape, type, buffer=None, byteoffset=0,
                 bytestride=None, byteorder=_sys.byteorder, aligned=1):

        # need to add a whole bunch of checking here (or then again,
        # maybe not.)

        type = _getType(type)
        itemsize = type.bytes
        NDArray.__init__(self, shape, itemsize,
                         buffer=buffer,
                         byteoffset=byteoffset,
                         bytestride=bytestride)
        self._type = type
        if byteorder in ["little", "big"]:
            self._byteorder = byteorder
        else:
            raise ValueError("byteorder must be 'little' or 'big'")
        # This attribute is used by subclasses to indicate to ufunc that
        # they should be called to perform the ufunc (e.g., complex numbers).
        self._myufuncs = None

    def __getstate__(self):
        """returns state of NumArray for pickling."""
        assert not hasattr(self, "_shadows") # Not a good idea for pickling.
        d = NDArray.__getstate__(self)
        return d

    def __setstate__(self, state):
        """sets state of NumArray after unpickling."""
        NDArray.__setstate__(self, state)
    
    def __del__(self): # Copied to __del__!
        """This method supports the use of temporary shadow arrays by the
        C-API.  The temporary refers to the original misbehaved array via
        the '_shadows' attribute.  When the temporary is deallocated, it
        copies itself back onto the original misbehaved array.
        
        Here, misbehaved means 'hard to work with in C', i.e., byteswapped,
        misaligned, etc.
        """
        if hasattr(self, "_shadows"):
            self._shadows._copyFrom(self);

    def isbyteswapped(self):
        """True if data in buffer is byteswapped relative to native form"""
        return self._byteorder != sys.byteorder and self._itemsize > 1

    def byteswap(self):
        """Byteswap the data in place.  Does *not* change _byteorder."""
        if self._itemsize != 1:
            fname = "byteswap"+str(self._itemsize)+"bytes"
            cfunc = _bytes.functionDict[fname]
            cfunc(self._shape, 
                  self._data, self._byteoffset, self._strides,
                  self._data, self._byteoffset, self._strides)
        
    def _togglebyteorder(self):
        "reverses the state of the _byteorder attribute:  little <-> big."""
        self._byteorder = {"little":"big","big":"little"}[self._byteorder]
        
    def info(self):
        """info() prints out the key attributes of a numarray."""
        ndarray.NDArray.info(self)
        print "byteorder:", self._byteorder
        print "byteswap:", self.isbyteswapped()
        print "type:", repr(self._type)

    def astype(self, type=None):
        """Return a copy of the array converted to the given type"""
        if type is None:
            type = self._type
        type = _getType(type)
        if type == self._type:
            # always return a copy even if type is same
            return self.copy()
        if type._conv:
            retarr = NewArray(buffer=None, shape=self._shape, type=type)
            if self.is_c_array():
                _numarray.CheckFPErrors()
                cfunc = self._type._conv.astype[type.name]
                cfunc(self.nelements(), 1, 1,
                      ((self._data, 0), (retarr._data, 0)))
                errorstatus = _numarray.CheckFPErrors()
                if errorstatus:
                    handleError(errorstatus, " during type conversion")
            else:
                ufunc._copyFromAndConvert(self, retarr)
        elif type.fromtype:
            retarr = type.fromtype(self)
        else:
            raise TypeError("Don't know how to convert from %s to %s" %
                            (self._type.name, type.name))
        return retarr

    def is_c_array(self):
        """returns 1 iff an array is contigous, aligned, and not
        byteswapped."""
        return (self.iscontiguous() and self.isaligned() and not
                self.isbyteswapped())

    def new(self, type):
        """Return a new array of given type with same shape as this array

        Note this only creates the array; it does not copy the data.
        """
        # return NewArray(buffer=None, shape=self._shape, type=type)
        if _isComplexType(type):
            return ComplexArray(buffer=None, shape=self._shape, type=type)
        else:
            return NumArray(buffer=None, shape=self._shape, type=type)
        
    def type(self):
        """Return the type object for the array"""
        return self._type

    def tolist(self):
        """Convert array to list(s) of Python scalars"""
        return array2list(self)

    def _getitem(self, key):
        """Returns value for single element in array using byte offset

        Does no validity checking on key (assumes it is valid).
        """
        byteoffset = self._getByteOffset(key)
        return self._type._conv.asPyValue(self._data, byteoffset,
                                          self._itemsize, self.isbyteswapped())

    def _setitem(self, key, value):
        """Sets a single value in an array using byte offset

        Does no validity checking on key (assumes it is valid).
        """
        byteoffset = self._getByteOffset(key)
        self._type._conv.fromPyValue(value, self._data, byteoffset, 
                                     self._itemsize, self.isbyteswapped())

    def _copyFrom(self, arr):
        """Copy elements from another array.

        This overrides the generic NDArray version
        """
        # Test for simple case first
        if isinstance(arr, NumArray):
            if (arr._type == self._type and
                (self._shape in [(1,), (), arr._shape] and
                 arr._shape in [(1,), (), self._shape]) and
                arr._byteorder == self._byteorder and
                ndarray.product(arr._strides) != 0 and
                arr.isaligned() and self.isaligned()):
                
                sShape = self._shape
                if arr._shape == ():
                    sShape = ()
                    
                name = 'copy'+`self._itemsize`+'bytes'
                cfunc = ( _bytes.functionDict.get(name) or
                          _bytes.functionDict["copyNbytes"])
                cfunc(sShape, arr._data,  arr._byteoffset,  arr._strides,
                      self._data, self._byteoffset, self._strides)
                
        elif PyNUMERIC_TYPES.has_key(_type(arr)):
            # Convert scalar to a one element array for broadcasting
            arr = array([arr])
        else:
            raise TypeError('argument is not array or number')
        barr = self._broadcast(arr, raiseOnFail=1)
        ufunc._copyFromAndConvert(barr, self)

    def copy(self):
        c = ndarray.NDArray.copy(self)
        if self.isbyteswapped():
            c.byteswap()
            c._togglebyteorder()
        return c
 
    def __str__(self):
        return arrayprint.array2string(self,
                MAX_LINE_WIDTH, PRECISION, SUPPRESS_SMALL, ' ', 0)

    def __repr__(self):
        return arrayprint.array2string(self,
                MAX_LINE_WIDTH, PRECISION, SUPPRESS_SMALL, ', ', 1)

    def __neg__(self): return ufunc.minus(self)
    def __add__(self, operand): return ufunc.add(self, operand)
    def __radd__(self, operand): return ufunc.add(operand, self)
    def __sub__(self, operand): return ufunc.subtract(self, operand)
    def __rsub__(self, operand): return ufunc.subtract(operand, self)
    def __mul__(self, operand): return ufunc.multiply(self, operand)
    def __rmul__(self, operand): return ufunc.multiply(operand, self)
    def __div__(self, operand): return ufunc.divide(self, operand)
    def __rdiv__(self, operand): return ufunc.divide(operand, self)
    def __mod__(self, operand): return ufunc.remainder(self, operand)
    def __rmod__(self, operand): return ufunc.remainder(operand, self)
    def __pow__(self, operand): return ufunc.power(self, operand)
    def __rpow__(self, operand): return ufunc.power(operand, self)

    def __and__(self, operand): return ufunc.bitwise_and(self, operand)
    def __rand__(self, operand): return ufunc.bitwise_and(operand, self)
    def __or__(self, operand): return ufunc.bitwise_or(self, operand)
    def __ror__(self, operand): return ufunc.bitwise_or(operand, self)
    def __xor__(self, operand): return ufunc.bitwise_xor(self, operand)
    def __rxor__(self, operand): return ufunc.bitwise_xor(operand, self)
    def __invert__(self): return ufunc.bitwise_not(self)
    def __rshift__(self, operand): return ufunc.rshift(self, operand)
    def __rrshift__(self, operand): return ufunc.rshift(operand, self)
    def __lshift__(self, operand): return ufunc.lshift(self, operand)
    def __rlshift__(self, operand): return ufunc.lshift(operand, self)
    def __pos__(self): return self

    # augmented assignment operators

    def __iadd__(self, operand):
        ufunc.add(self, operand, self)
        return self
    
    def __isub__(self, operand):
        ufunc.subtract(self, operand, self)
        return self
    
    def __imul__(self, operand):
        ufunc.multiply(self, operand, self)
        return self
    
    def __idiv__(self, operand):
        ufunc.divide(self, operand, self)
        return self
    
    def __imod__(self, operand):
        ufunc.remainder(self, operand, self)
        return self
    
    def __ipow__(self, operand):
        ufunc.power(self, operand, self)
        return self
    
    def __iand__(self, operand):
        ufunc.bitwise_and(self, operand, self)
        return self
    
    def __ior__(self, operand):
        ufunc.bitwise_or(self, operand, self)
        return self
    
    def __ixor__(self, operand):
        ufunc.bitwise_xor(self, operand, self)
        return self
    
    def __irshift(self, operand):
        ufunc.rshift(self, operand, self)
        return self
    
    def __ilshift(self, operand):
        ufunc.lshift(self, operand, self)
        return self

    # rich comparisons (only works in Python 2.1 and later)

    def __lt__(self, operand): return ufunc.less(self, operand)
    def __gt__(self, operand): return ufunc.greater(self, operand)
    def __le__(self, operand): return ufunc.less_equal(self, operand)
    def __ge__(self, operand): return ufunc.greater_equal(self, operand)
    def __eq__(self, operand):
        if operand is None:
            return 0
        else:
            return ufunc.equal(self, operand)
    def __ne__(self, operand):
        if operand is None:
            return 1
        else:
            return ufunc.not_equal(self, operand)
    def __nonzero__(self):
        b = self.view()
        b.ravel()
        return bitwise_or.reduce(b.__ne__(0))

    def sort(self, axis=-1):
        if axis==-1:
            ufunc._sortN(self)
        else:
            self.swapaxes(axis,-1)
            ufunc._sortN(self)
            self.swapaxes(axis,-1)

    def _argsort(self, axis=-1):
        if axis==-1:
            ashape = self.getshape()
            w = array(shape=ashape, type=Long)
            w[...,:] = arange(ashape[-1], type=Long)
            ufunc._argsortN(self,w)
            return w
        else:
            self.swapaxes(axis,-1)
            return self._argsort()

    def argsort(self, axis=-1):
        return self.copy()._argsort(axis)

    def argmax(self, axis=-1):
        return self.argsort(axis)[...,-1]

    def argmin(self, axis=-1):
        return self.argsort(axis)[...,0]

    def diagonal(self, *args, **keywords):
        return diagonal(self, *args, **keywords)

    def trace(self, *args, **keywords): 
        return trace(self, *args, **keywords)

    def typecode(self):
        return numerictypes.typecode[self._type]


class ComplexArray(NumArray):
    """Complex Arrays

    Although it inherits from Numarray, this class inherits virtually
    none of its methods! The inheritance is solely so that complex
    arrays can be tested as an instance of NumArrays and thus be
    properly considered as such.
    """

    def __init__(self, shape, type, buffer=None, byteoffset=0, bytestride=None,
            byteorder=_sys.byteorder, aligned=1, real=None, imag=None):

        type = _getType(type)
        if not _isComplexType(type):
            raise  TypeError("Type must be a Complex number type")
        itemsize = type.bytes
        NumArray.__init__(self, shape, type, buffer=buffer,
                          byteoffset=byteoffset,
                          bytestride=bytestride,
                          byteorder=byteorder)
        if not bytestride:
            rbytestride = itemsize
        else:
            rbytestride = bytestride
        if real is not None:
            self.getreal()[:] = real
        if imag is not None:
            self.getimag()[:] = imag

    def byteswap(self):
        """Byteswap the data in place; customized for Complex."""
        if self._itemsize != 1:
            fname = "byteswap" + self._type.name
            cfunc = _bytes.functionDict[fname]
            cfunc(self._shape, 
                  self._data, self._byteoffset, self._strides,
                  self._data, self._byteoffset, self._strides)
        
    def itemsize(self):
        return 2*ndarray.NDArray.itemsize(self)

    def getreal(self):
        t = _realtype(self._type)
        arr = NumArray(self._shape, t, buffer=self._data,
                       byteoffset=self._byteoffset,
                       bytestride=self._bytestride,
                       byteorder=self._byteorder)
        arr._strides = self._strides[:]
        return arr

    def setreal(self, value):
        self.getreal()[:] = value
        
    def getimag(self):
        t = _realtype(self._type)
        arr = NumArray(self._shape, t, buffer=self._data,
                       byteoffset=self._byteoffset+t.bytes,
                       bytestride=self._bytestride,
                       byteorder=self._byteorder)
        arr._strides = self._strides[:]
        return arr

    def setimag(self, value):
        self.getimag()[:] = value

    imag = property(getimag, setimag,
                        doc="imaginary component of complex array")

    real = property(getreal, setreal,
                        doc="real component of complex array")

    setimaginary = setimag
    getimaginary = getimag
    imaginary = imag

    def _getitem(self, key):
        """Returns value for single element in array using byte offset

        Does no validity checking on byteoffset (assumes it is valid).
        """
        return complex(self.getreal()._getitem(key),
                       self.getimag()._getitem(key))

    def _setitem(self, key, value):
        """Sets a single value in an array using byte offset

        Does no validity checking on byteoffset (assumes it is valid).
        """
        cvalue = complex(value)
        self.getreal()._setitem(key, cvalue.real)
        self.getimag()._setitem(key, cvalue.imag)

    def _copyFrom(self, arr):
        """Copy elements from another array.

        This overrides the NumArray version
        """
        if   isinstance(arr, ComplexArray):
            real, imag = arr.getreal(), arr.getimag()
        elif _type(arr)==types.ComplexType:
            real, imag = arr.real,  arr.imag
        else:
            real, imag = arr,       0.
        self.getreal()._copyFrom(real)
        self.getimag()._copyFrom(imag)

    def conjugate(self):
        ufunc.minus(self.getimag(), self.getimag())

    def dummy():
        """
arctan2
bitwise_and",
bitwise_or",
bitwise_xor",
bitwise_not",
rshift",  2,
lshift",  2,
"""
        pass

def Complex32_fromtype(arr):
    """Used for converting other types to Complex32.

    This is used to set an fromtype attribute of the ComplexType objects"""
    rarr = arr.astype(Float32)
    retarr = ComplexArray(arr._shape, Complex32)
    retarr.getreal()[:] = rarr
    retarr.getimag()[:] = 0.
    return retarr

def Complex64_fromtype(arr):
    """Used for converting other types to Complex64.

    This is used to set an fromtype attribute of the ComplexType objects"""
    rarr = arr.astype(Float64)
    retarr = ComplexArray(arr._shape, Complex64)
    retarr.getreal()[:] = rarr
    retarr.getimag()[:] = 0.
    return retarr

# Check whether byte order is big endian or little endian.

from sys import byteorder
isBigEndian = (byteorder == "big")
del byteorder

# Add fromtype function to Complex types

Complex32.fromtype = Complex32_fromtype
Complex64.fromtype = Complex64_fromtype

# Return type of complex type components

def _isComplexType(type):
    return type in [Complex32, Complex64]

def _realtype(complextype):
    if complextype == Complex32:
        return Float32
    else:
        return Float64


def conjugate(a):
    """conjugate(a) returns the complex conjugate of 'a'"""
    a = inputarray(a)
    if not _isComplexType(a._type):
        a = a.astype(Complex64)
    a.conjugate()
    return a



def zeros(shape, type=None):
    shape = ndarray.getShape(shape)
    retarr = _fillarray(ndarray.product(shape), 0, 0, type)
    retarr.setshape(shape)
    return retarr

def ones(shape, type=None):
    shape = ndarray.getShape(shape)
    retarr = _fillarray(ndarray.product(shape), 1, 0, type)
    retarr.setshape(shape)
    return retarr

def arange(a1, a2=None, stride=1, type=None, shape=None):
    if a2 == None:
        start = 0 + (0*a1) # to make it same type as stop
        stop  = a1
    else:
        start = a1 +(0*a2)
        stop  = a2
    delta = (stop-start)/stride ## xxx int divide issue
    if _type(delta) == types.ComplexType:
        # xxx What make sense here?
        size = int(math.ceil(delta.real))
    else:
        size = int(math.ceil((stop-start)/float(stride)))
    r = _fillarray(size, start, stride, type)
    if shape is not None:
        r.setshape(shape)
    return r

arrayrange = arange  # alias arange as arrayrange.

def identity(n, type=None):
    a = zeros(shape=(n,n), type=type)
    i = arange(n)
    a.put(i, i, 1)
    return a

def dot(a,b):
    """dot(a,b) matrix-multiplies a by b.
    """
    return innerproduct(a, swapaxes(inputarray(b), -1, -2))

matrixmultiply = dot  # Deprecated in Numeric

def outerproduct(a,b):
    """outerproduct(a,b) computes the NxM outerproduct of N vector 'a' and
    M vector 'b', where result[i,j] = a[i]*b[j].
    """
    a=reshape(inputarray(a), (-1,1))  # ravel a into an Nx1
    b=reshape(inputarray(b), (1,-1))  # ravel b into a 1xM
    return matrixmultiply(a,b)   # return NxM result

def sum(a, axis=0):
    """sum(a,axis=0) is a synonym for add.reduce(a, axis).
    """
    a = inputarray(a)
    if axis == 0:
        return ufunc.add.reduce(a)
    else:
        return swapaxes(ufunc.add.reduce(swapaxes(a, 0, axis)), 0, axis)

def cumsum(a, axis=0):
    """cumsum(a,axis=0) is a synonym for add.accumulate(a, axis).
    """
    a = inputarray(a)
    if axis == 0:
        return ufunc.add.accumulate(a)
    else:
        return swapaxes(ufunc.add.accumulate(swapaxes(a, 0, axis)), 0, axis)

def product(a, axis=0):
    """product(a,axis=0) is a synonym for multiply.reduce(a, axis).
    """
    a = inputarray(a)
    if axis == 0:
        return ufunc.multiply.reduce(a)
    else:
        return swapaxes(ufunc.multiply.reduce(swapaxes(a, 0, axis)),
                        0, axis)

def cumproduct(a, axis=0):
    """cumproduct(a,axis=0) is a synonym for multiply.accumulate(a,axis).
    """
    a = inputarray(a)
    if axis == 0:
        return ufunc.multiply.accumulate(a)
    else:
        return swapaxes(ufunc.multiply.accumulate(swapaxes(a, 0, axis)),
                        0, axis)

def alltrue(a, axis=0):
    """alltrue(a,axis=0) is a synonym for logical_and.reduce(a,axis).
    """
    a = inputarray(a)
    if axis == 0:
        return ufunc.logical_and.reduce(a)
    else:
        return swapaxes(ufunc.logical_and.reduce(swapaxes(a, 0, axis)),
                        0, axis)
def sometrue(a, axis=0):
    """sometrue(a,axis=0) is a synonym for logical_or.reduce(a,axis).
    """
    a = inputarray(a)
    if axis == 0:
        return ufunc.logical_or.reduce(a)
    else:
        return swapaxes(ufunc.logical_or.reduce(swapaxes(a, 0, axis)),
                        0, axis)

def allclose (a, b, rtol=1.e-5, atol=1.e-8): # From Numeric 20.3
    """ allclose(a,b,rtol=1.e-5,atol=1.e-8)
        Returns true if all components of a and b are equal
        subject to given tolerances.
        The relative error rtol must be positive and << 1.0
        The absolute error atol comes into play for those elements
        of y that are very small or zero; it says how small x must be also.
    """
    x, y = inputarray(a), inputarray(b)
    d = less(abs(x-y), atol + rtol * abs(y))
    return alltrue(ndarray.ravel(d))


# From Numeric-21.0
def diagonal(a, offset= 0, axis1=0, axis2=1):
    """diagonal(a, offset=0, axis1=0, axis2=1) returns the given diagonals
    defined by the last two dimensions of the array.
    """
    a = inputarray(a)
    if axis2 < axis1: axis1, axis2 = axis2, axis1
    if axis2 > 1:
        new_axes = range (len (a._shape))
        del new_axes [axis2]; del new_axes [axis1]
        new_axes [0:0] = [axis1, axis2]
        a = ndarray.transpose (a, new_axes)
    s = a._shape
    if len (s) == 2:
        n1 = s [0]
        n2 = s [1]
        n = n1 * n2
        s = (n,)
        a = reshape (a, s)
        if offset < 0:
            return ndarray.take (a, range(- n2 * offset, min(n2, n1+offset) * (n2+1) - n2 * offset, n2+1), axis=0)
        else:
            return ndarray.take (a, range (offset,         min(n1, n2-offset) * (n2+1) + offset,      n2+1), axis=0)
    else :
        my_diagonal = []
        for i in range (s [0]) :
            my_diagonal.append (diagonal (a [i], offset))
        return array(my_diagonal)

# From Numeric-21.0
def trace(a, offset=0, axis1=0, axis2=1):
    """trace(a,offset=0, axis1=0, axis2=1) returns the sum along diagonals
    (defined by the last two dimenions) of the array.
    """
    return add.reduce(diagonal(a, offset, axis1, axis2))

def rank(a):
    """rank(a) returns the number of dimensions in 'a'"""
    return inputarray(a).getrank()

def shape(a):
    """rank(a) returns the shape tuple of 'a'"""
    return inputarray(a).getshape()

def size(a):
    """size(a) returns the size in bytes of array 'a'"""
    return inputarray(a).nelements()

def array_str(a):
    return str(inputarray(a))

def array_repr(a):
    return repr(inputarray(a))

def around(a, digits=0):
    a = inputarray(a)
    scale = 10.0 ** digits
    a = a / scale
    round(a,a)
    return a * scale

absolute = abs   # synonym
negative = minus # synonym
fmod = remainder # synonym

def explicit_type(x):
    """explicit_type(x) returns a view of x which will always show it's type in it's repr.
    This is useful when the same test is run in two places, one where the type used *is* the
    default and hence not normally repr'ed, and one where the type used *is not* the default
    and so is displayed.
    """
    y = x.view()
    y._explicit_type = 1
    return y
