"""NumArray Universal Function Module

$Id: ufunc.py,v 1.26 2002/06/17 13:56:24 jaytmiller Exp $
"""

import types
import memory
import _sort, _bytes
from operator import and_  # seems there are lots of symbols named 'operator'
import _ufunc
import ndarray
import _numarray

# note more imports are at end of file

#import pywin.debugger

def ufuncFactory(operator, ucfuncs, identity=None):

    """Create UFunc instance based on signature of functions"""

    ninputs, noutputs = _nIOArgs(ucfuncs)
    if ninputs==1 and noutputs==1:
        return _UnaryUFunc(operator, ucfuncs, identity)
    elif ninputs==2 and noutputs==1:
        return _BinaryUFunc(operator, ucfuncs, identity)
    elif noutputs==1:
        raise ValueError("don't handle ufuncs with more than 2 input arguments yet")
    else:
        raise ValueError("don't handle ufuncs with more than 1 output argument yet")

def _nIOArgs(ucfuncs):

    """Determine number of arguments from signature

    Returns tuple of number of input and output arguments.
    Also check for consistency in numbers of arguments.
    """

    # The number of items in the optype key implies the number of inputs,
    # hence the number of inputs = # of underscores+1
    noptypeinputs = []
    ninputs = []
    noutputs = []
    for key in ucfuncs.keys():
        if key in ['reduce', 'accumulate']:
            continue
        noptypeinputs.append(key.count('_'))
        for insigkey in ucfuncs[key].keys():
            ninputs.append(len(insigkey))
            noutputs.append(len(ucfuncs[key][insigkey][0]))
    if max(noptypeinputs) != min(noptypeinputs):
        raise StandardError(
            "inconsistency in operator types for operator: "+operator)
    if max(ninputs) != min(ninputs):
        raise StandardError(
            "not all input signatures match in number of arguments\n" +
            "for operator: "+operator)
    if max(noutputs) != min(noutputs):
        raise StandardError(
            "not all output signatures match in number of arguments\n" +
            "for operator: "+operator)
    return max(ninputs), max(noutputs)


class _BufferPool:
    """A Pool of available buffers available for block compuations

    Buffers can be gotten. When they are deleted they are automatically
    returned to the buffer pool so they can be reused.
    """
    def __init__(self, buffersize=10000):
        # buffers are initially allocated as needed
        self.buffers = []
        self.buffersize = buffersize

    def getBuffer(self):
        try:
            return _UBuffer(self.buffers.pop())
        except IndexError:
            # out of buffers, create a new one
            return _UBuffer(memory.new_memory(self.buffersize))

    def setBufferSize(self, buffersize):
        """Change buffer size (useful mainly for testing)"""
        if buffersize != self.buffersize:
            self.buffersize = buffersize
            # delete existing (wrong-sized) buffers
            self.buffers = []


class _UBuffer:

    """Wrapper around buffer object

    A feeble attempt to share all relevant attributes with an array object
    (namely ._data, ._strides, ...)
    """

    def __init__(self, pybuffer):
        self._data = pybuffer
        self._byteoffset = 0
        self._strides = None        # how it is distinguished from a real array

    def isbyteswapped(self): return 0
    def isaligned(self): return 1
    def iscontiguous(self): return 1
    def is_c_array(self): return 1

    def _getByteOffset(self, indices): return 0

    def __del__(self):
        """On deletion return the data to the buffer pool"""
        if self._data is not None:
            _bufferPool.buffers.append(self._data)
            self._data = None # Lose the reference to the buffer.

_blockingParametersCache = {}

def _getBlockingParameters(shape, niter, overlap=0):
    """Figure out how to split the input array into subarrays
    close (but less than) the size of the block buffers.
    overlap is used in cases where it is necessary to overlap the
    last dimension in computations (for example for reduce and
    accumulate.)

    returns:

    level:      index at which the c functions should be called
    blockingparameters:  tuple consisting of
         number of full subarrays at the given level
         subarray shape
         a flag indicating if there is a leftover part of the array
             which is smaller than the regular subarray (leftover)
         the shape of the leftover subarray
    """
    retval = _blockingParametersCache.get((shape,niter,overlap))
    if retval is not None:
        return retval
    
    if shape == ():
        retval = 0, (1, (), 0, None)
        _blockingParametersCache[(shape,niter,overlap)] = retval
        return retval
    
    ndim = len(shape)
    csize = long(shape[-1]) # cumulative shape
    if csize > niter:
        # last dimension is too big for buffer
        # special overlap handling in this case
        nblocks, leftover = divmod(csize, niter-overlap)
        nblocks, leftover = int(nblocks), int(leftover)
        if leftover:
            retval = ndim-1, (nblocks, (niter,), 1, (leftover,))
        else:
            retval = ndim-1, (nblocks, (niter,), 0, None)
    else:
        for i in xrange(2,ndim+1):
            csize *= shape[-i]
            if csize >= niter:
                # need to break up this dimension
                tdim = long(shape[-i])
                blockdim = niter*tdim/csize
                nblocks, leftover = divmod(tdim, blockdim)
                nblocks, leftover = int(nblocks), int(leftover)
                stail = shape[-i+1:]
                if leftover:
                    retval = ndim-i, (nblocks, (blockdim,)+stail,
                        1, (leftover,) + stail)
                else:
                    retval = ndim-i, (nblocks, (blockdim,)+stail, 0, None)
                break
        else:
            # don't need to break up
            retval = 0, (1, shape, 0, None)
    _blockingParametersCache[(shape,niter,overlap)] = retval
    return retval

def _doOverDimensions(functions, outshape,
            dims, indexlevel, blockingparameters, overlap=0, level=0):
    """This version handles alignment, byteswapping and n-dimensions

    This is a recursive method
    dims starts out as an empty list, and accumulates the indices
    for each dimension. Eg. if the array has shape (7, 3, 5) on
    the initial call, the next level call will specify the the
    'first' index (corresponding to the 7), the next will have
    the first two indices (and that would be the last level since
    the lowest dimension is taken care of by the cfuncs.)
    """
    if level == indexlevel:
        nregShapeIters, shape, leftover, leftoverShape, = blockingparameters
        if shape != ():
            for i in xrange(nregShapeIters):
                for f in functions:
                    f(dims+[i*(shape[0]-overlap)], shape)
            if leftover:
                for f in functions:
                    f(dims+[(i+1)*(shape[0]-overlap)], leftoverShape)
        else:
                for f in functions:
                    f([0], shape)
    else:
        # recurse
        for i in xrange(outshape[level]):
            _doOverDimensions(functions, outshape,
                              dims+[i], indexlevel, blockingparameters,
                              overlap, level+1)

def _copyFromAndConvert(inarr, outarr):

    """Copy from one array to another handling strides, form & conv"""

    if inarr._shape != outarr._shape:
        raise ValueError("Arrays must have the same shape")
    # Type conversion done using inputconverter
    input = _InputConverter(inarr, outarr._type.name)
    # Ensure output step uses output buffer of input step
    output = _OutputConverter(outarr, outarr._type.name,
                              inbuffer=input.outbuffer,
                              forcestride=1)
    maxitemsize = max(inarr._type.bytes, outarr._type.bytes)
    niter = _bufferPool.buffersize/maxitemsize
    indexlevel, blockingparameters = \
                _getBlockingParameters(outarr._shape, niter)
    functions = (input.compute, output.compute)
    _numarray.CheckFPErrors()
    _doOverDimensions(functions, outarr._shape, [],
                           indexlevel, blockingparameters)
    errorstatus = _numarray.CheckFPErrors()
    if errorstatus:
        numarray.handleError(errorstatus, " during type conversion")

def _noConversionNeeded(x, xt):
    """returns 1 iff x is a c_array of type xt."""
    return x.is_c_array() and (x._type.name == xt or x._type == xt)

def _inputcheck(*inargs):

    """Check input args for type, convert sequences to numarrays

    Return list of 1) input arg tuple with scalars and numarrays only
                   2) signature string indicating which args are scalars
                      and which are numarrays
                   3) flag indicating if only scalar arguments are present
    """

    retargs = []
    retsig = []
    scalar = 1
    for inarg in inargs:
        if isinstance(inarg, numarray.NumArray):
            scalar = 0
            retsig.append('vector')
        elif numarray.PyNUMERIC_TYPES.has_key(type(inarg)):
            retsig.append('scalar')
        else:
            scalar = 0
            retsig.append('vector')
            # see if it can be made into an array
            try:
                inarg = numarray.array(inarg)
            except TypeError:
                raise TypeError(
                 "UFunc arguments must be arrays, scalars or numeric sequences")
        retargs.append(inarg)
    return retargs, '_'.join(retsig), scalar

def _pseudoArrayFromScalars(scalarvalues, type):
    """Wrap a scalar in a buffer so it can be used as an array"""
    arr = _bufferPool.getBuffer()
    newtype = _numtypedict[type]
    for i in xrange(len(scalarvalues)):
        s = scalarvalues[i];
        newtype._conv.fromPyValue(s, arr._data, i*newtype.bytes,
                                  newtype.bytes, 0)
    #  Modify block buffer attributes to look like vector/vector setup.
    arr._type = newtype
    arr._shape = (len(scalarvalues),)
    arr._itemsize = newtype.bytes
    return arr

def callArrayUnaryUFuncMethod(ufunc, inarr, outarr):
    """Perform Ufunc by calling array method"""
    methodname = "_"+ufunc.operator
    ufmethod = getattr(inarr, methodname)
    return ufmethod(outarr)

def callArrayBinaryUFuncMethod(ufunc, inarr1, inarr2, outarr):
    """Perform Ufunc by calling array method"""
    methodname = "_"+ufunc.operator
    ufmethod = getattr(inarr1, methodname)
    return ufmethod(inarr2, outarr)

def callArrayBinaryUFuncRightMethod(ufunc, inarr1, inarr2, outarr):
    """Perform Ufunc by calling array method"""
    methodname = "_right_"+ufunc.operator
    ufmethod = getattr(inarr2, methodname)
    return ufmethod(inarr1, outarr)

class _UFunc:
    """Class to handle all element-by-element mathematical functions

    This is the base class -- subclasses must implement __call__ and
    (if allowed) the outer and _cumulative methods.
    """

    def __init__(self, operator, ufuncs, identity):
        self.operator = operator
        self._ucfuncs = ufuncs
        self._identity = identity
        self._coercions = numerictypes.genericCoercions
        self._promotionExclusions = numerictypes.genericPromotionExclusions
        self._typerank = numerictypes.genericTypeRank

    def _getidentity(self):
        i = numarray.array([self._identity])
        i.setshape(())
        return i

    def __call__(self, *args):
        raise NotImplementedError("__call__ is not implemented by base UFunc class")

    def _doit(self, computation_mode, outarr, cfunc, ufargs, scalar):
        _numarray.CheckFPErrors()
        if computation_mode == "fast":
            apply(cfunc, ufargs)
        else:
            inputs, outputs, preprocessing_outbuffers, \
            postprocessing_inbuffers, maxitemsize = ufargs
            niter = _bufferPool.buffersize/maxitemsize

            outshape = outarr._shape
            indexlevel, blockingparameters = \
                _getBlockingParameters(outshape, niter)
            operator = _Operator(cfunc, preprocessing_outbuffers,
                                 postprocessing_inbuffers)
            functions = ([ f.compute for f in inputs] +
                         [ operator.compute ] +
                         [ f.compute for f in outputs])
            _doOverDimensions(functions, outshape, [],
                indexlevel, blockingparameters)
        errorstatus = _numarray.CheckFPErrors()
        if errorstatus:
            numarray.handleError(errorstatus, " in "+self.operator)
        if scalar:
            return outarr[0]
        else:
            return outarr

    def _typematch(self, inputSignature, bintypeDict, scalar=0):

        """Determine which C function is most appropriate for the given types

        returns the input types the inputs must be converted to, the
        output type, and the cfunc object
        """

        # First see if there is an exact match
        if bintypeDict.has_key(inputSignature):
            return inputSignature + bintypeDict[inputSignature]
        else:
            # Look for 'closest' diagonal element.
            # First find 'lowest' common type.
            intype = self._coercions[inputSignature]
            keys = bintypeDict.keys()
            atypelist = []
            for key in keys:
                if key[0] == key[1]:
                    atypelist.append(key[0])
            newInputSignature = (self._typePromoter(intype, atypelist),)*2
            return newInputSignature + bintypeDict[newInputSignature]

    def _typePromoter(self, intype, typelist):

        """Type a given type should be promoted to given a list of choices"""

        # type as string for now
        # Not an efficient version for now
        # Generate the type specific rank list
        tranks = self._typerank[:] # copy since we will be deleting elements
        exclusions = self._promotionExclusions[intype]
        for exclusion in exclusions:
            del tranks[exclusion]
        index = tranks.index(intype)
        tranks = tranks[index:]
        # Which one has lowest rank?
        minimumRank = 100
        for atype in typelist:
            try:
                rank = tranks.index(atype)
                if rank < minimumRank:
                    minimumRank = rank
            except ValueError:
                pass
        if minimumRank < 100:
            return tranks[minimumRank]
        else:
            raise TypeError("unable to find type to promote to")

    def _typeDemoter(self, intype, typelist):

        """The opposite of what typePromoter does--not implemented yet"""

        pass

    def outer(self, inarr1, inarr2, outarr=None):
        raise ValueError("Outer only available for binary functions")

class _UnaryUFunc(_UFunc):

    """Class for ufuncs with 1 input and 1 output argument"""

    def __call__(self, inarr, outarr=None):

        """The standard calling interface for a unary UFunc"""

        # See if array handles its own ufuncs (complex for example)
        (inarr,), insig, scalar = _inputcheck(inarr)
        if isinstance(inarr, numarray.NumArray) and inarr._myufuncs:
            return callArrayUnaryUFuncMethod(self, inarr, outarr)
        else:
            computation_mode, woutarr, cfunc, ufargs = \
                              self._setup(inarr, insig, outarr)
            result = self._doit(computation_mode, woutarr,
                                cfunc, ufargs, scalar)
            if outarr is None:
                return result
            else:
                return None

    def _setup(self, inarr, insig, outarr):
        """Setup for unary ufunc"""
        ufdict = self._ucfuncs["vector"]
        if insig == "scalar":
            inarr =  numarray.array([inarr])
            if outarr is not None:
                raise ValueError("Error to supply output array with no array arguments")
        shape = inarr._shape
        typekeys  = [key[0] for key in ufdict.keys()]
        intype = self._typePromoter(inarr._type.name, typekeys)
        outtype, ucfunc = ufdict[(intype,)]
        if outarr is None:
            outarr = inarr.new(_numtypedict[outtype[0]])
        if shape != outarr._shape:
            raise ValueError("Supplied output array does not have appropriate shape")
        # If fast case, return the arguments for an immediate call to the
        #  C function
        if (_noConversionNeeded(inarr, intype) and
            _noConversionNeeded(outarr, outtype[0])):
            return ("fast", outarr, ucfunc, (outarr.nelements(), 1, 1, 
                ((inarr._data, inarr._byteoffset),
                (outarr._data, outarr._byteoffset))))
        # slower, general case
        inputs = (_InputConverter(inarr, intype), )
        outputs = (_OutputConverter(outarr=outarr, intype=outtype[0]),)
        # find largest itemsize involved, it will determine how many
        # iterations can be done on the temporary buffers (i.e., niter)
        maxitemsize = max(_numtypedict[inarr._type.name].bytes,
                          _numtypedict[intype].bytes,
                          _numtypedict[outtype[0]].bytes,
                          _numtypedict[outarr._type.name].bytes)
        preprocessing_outbuffers = (inputs[0].outbuffer,)
        postprocessing_inbuffers = (outputs[0].inbuffer,)
        return ("slow", outarr, ucfunc, 
            (inputs, outputs, preprocessing_outbuffers,
             postprocessing_inbuffers, maxitemsize))

class _BinaryUFunc(_UFunc):

    """Class for ufuncs with 2 input and 1 output arguments"""

    def __call__(self, inarr1, inarr2, outarr=None):

        """The standard calling interface for UFuncs"""
        # See if array handles its own ufuncs (complex for example)
        (inarr1, inarr2), insig, scalar = _inputcheck(inarr1, inarr2)
        if (isinstance(inarr1, numarray.NumArray) and inarr1._myufuncs):
            return callArrayBinaryUFuncMethod(self, inarr1, inarr2, outarr)
        elif (isinstance(inarr2, numarray.NumArray) and inarr2._myufuncs):
            return callArrayBinaryUFuncRightMethod(self, inarr1, inarr2, outarr)
        else:
            computation_mode, woutarr, cfunc, ufargs = \
                          self._setup(inarr1, inarr2, insig, outarr)
            result = self._doit(computation_mode, woutarr,
                                cfunc, ufargs, scalar)
            if outarr is None:
                return result
            else:
                return None

    def _setup(self, in1, in2, insig, outarr):

        """Setup for binary ufunc"""

        if insig == "scalar_scalar":
            # do it as two 1-d arrays.
            in1 = numarray.array([in1])
            in2 = numarray.array([in2])
            insig = "vector_vector"
            if outarr is not None:
                raise ValueError(
                    "Error to supply output array with no array arguments")
        elif insig == "vector_scalar":
            intypes = (in1._type.name, type(in2))
            convtype1, convtype2, outtype, ucfunc \
                = self._typematch(intypes, self._ucfuncs[insig])
            inarr1 = in1
            inarr2 = _pseudoArrayFromScalars([in2], convtype2)
            shape = inarr1._shape
            if outarr is None:
                outarr = inarr1.new(_numtypedict[outtype[0]])
        elif insig == "scalar_vector":
            intypes = (type(in1), in2._type.name)
            convtype1, convtype2, outtype, ucfunc \
                = self._typematch(intypes, self._ucfuncs[insig])
            inarr1 = _pseudoArrayFromScalars([in1], convtype1)
            inarr2 = in2
            shape = inarr2._shape
            if outarr is None:
                outarr = inarr2.new(_numtypedict[outtype[0]])
        if insig == "vector_vector":
            # See if one or the other arrays is shape consistent or can be broadcast
            inarr1, inarr2 = in1._dualbroadcast(in2, raiseOnFail=1)
            intypes = (in1._type.name, in2._type.name)
            convtype1, convtype2, outtype, ucfunc \
                = self._typematch(intypes, self._ucfuncs[insig])
            shape = inarr1._shape
            if outarr is None:
                outarr = inarr1.new(_numtypedict[outtype[0]])
        if shape != outarr._shape:
            raise ValueError("Supplied output array does not have appropriate shape")
        # If fast case, return the arguments for an immediate call to the
        #  C function
        if (_noConversionNeeded(inarr1, convtype1) and
            _noConversionNeeded(inarr2, convtype2) and
            _noConversionNeeded(outarr, outtype[0])):
            return ("fast", outarr, ucfunc, (outarr.nelements(), 2, 1, 
                ((inarr1._data, inarr1._byteoffset),
                (inarr2._data, inarr2._byteoffset),
                (outarr._data, outarr._byteoffset))))
        # slower, general case
        inputs = (_InputConverter(inarr1, convtype1),
                  _InputConverter(inarr2, convtype2))
        outputs = (_OutputConverter(outarr=outarr, intype=outtype[0]),)
        # find largest itemsize involved, it will determine how many
        # iterations can be done on the temporary buffers (i.e., niter)
        maxitemsize = max(_numtypedict[inarr1._type.name].bytes,
                          _numtypedict[inarr2._type.name].bytes,
                          _numtypedict[convtype1].bytes,
                          _numtypedict[convtype2].bytes,
                          _numtypedict[outtype[0]].bytes,
                          _numtypedict[outarr._type.name].bytes)
        preprocessing_outbuffers = (inputs[0].outbuffer, inputs[1].outbuffer)
        postprocessing_inbuffers = (outputs[0].inbuffer,)
        return ("slow", outarr, ucfunc, 
            (inputs, outputs, preprocessing_outbuffers,
             postprocessing_inbuffers, maxitemsize))

    def outer(self, inarr1, inarr2, outarr=None):

        """Return outer product of 2 arrays"""

        (inarr1, inarr2), insig, scalar = _inputcheck(inarr1, inarr2)
        ndim2 = len(inarr2._shape)
        indexarg = (slice(None, None, None),)*len(inarr1._shape) + (None,)*ndim2
        if outarr is not None:
            self(inarr1.__getitem__(indexarg), inarr2, outarr)
        else:
            return self(inarr1.__getitem__(indexarg), inarr2)

    _bad_shape = "Supplied output array does not have the appropriate shape"

    def _accumulate_out(self, inarr, outarr, outtype):
        """_accumulate_out creates temp output array for accumulate"""
        # Create output array if not supplied.
        if outarr is None:
            toutarr = inarr.new( outtype )
        else:
            if outarr._shape != inarr._shape:
                raise ValueError(self._bad_shape)
            toutarr = outarr
        # It is necessary to initialize the first subarray of the output
        # to the first subarray of the input (C functions depend on it).
        toutarr[...,0] = inarr[...,0]
        return toutarr

    def _reduce_out(self, inarr, outarr, outtype):
        # Create output array if not supplied.
        if outarr is None or outarr._type != outtype:
            toutarr = inarr[...,0].copy().astype(outtype)
            assert toutarr.is_c_array()
        else:  # outarr is defined and has type == outtype
            toutarr = outarr.view() # Get a view since we hack the strides.
            toutarr[...] = inarr[...,0]  # Assign first subarray of output
        toutarr._strides += (0,)
        return toutarr

    def _cumulative(self, cumoperation, inarr, outarr):

        """Used by both reduce and accumulate to compute accumulations.

        Assumes that the dimension of accumulation/reduction is 0.

            inarr        -- The input array to accumulate/reduce
            outarr       -- Optional output array

        The output array must have a consistent shape with the input
        Array (the same for accumulate and minus the accumulated dimension
        for reduce).

        The output type must match that produced by default
        and cannot be nonaligned or byteswapped. (It is a real mess to
        remove this restriction. (The original numeric doesn't even support
        output arrays.)
        """
        if (outarr is not None and
            (not outarr.isaligned() or outarr.isbyteswapped())):
            raise ValueError(
                "Reduce/Accumulate: no support for misaligned / byteswapped output arrays")

        if cumoperation not in ["reduce", "accumulate"]:
            raise ValueError("Unknown cumulative option")
        (inarr,), insig, scalar = _inputcheck(inarr)
        if scalar:
            raise ValueError("First argument must be an array or sequence")

        # The original numeric returns the operations identity for empty arrays.
        if inarr.nelements() == 0:
            if self._identity is not None:
                return self._getidentity() # 0D array form
            else:
                raise ValueError(
                    "Empty array supplied to a ufunc that has no identity value")

        sourcemsg = "in %s.%s" % (self.operator, cumoperation) # For error message

        # Get appropriate function, figure out any needed input type conversions
        # and what the output type will be
        ufdict = self._ucfuncs[cumoperation]
        typekeys  = [key[0] for key in ufdict.keys()]
        intype  = _numtypedict[self._typePromoter(inarr._type.name, typekeys)]
        outtype, ucfunc = ufdict[(intype,)]
        outtype = _numtypedict[outtype[0]]

        # A few words on shape/stride manipulations:
        # The C function assumes the last dimension is being accumulated.

        # Create the accumulate/reduce specific output array, toutarr
        #   type of reduce toutarr is always outtype to avoid output conversion
        #   with mismatched shapes.
        if cumoperation == "reduce":
            toutarr = self._reduce_out(inarr, outarr, outtype)
        else:
            toutarr = self._accumulate_out(inarr, outarr, outtype)

        # check for fast case (Unlike Ufuncs, contiguous arrays not required)
        if (not inarr.isbyteswapped() and inarr.isaligned()
            and intype == inarr._type and not toutarr.isbyteswapped()
            and toutarr.isaligned()):
            _numarray.CheckFPErrors()
            ucfunc(inarr._shape,
                   inarr._data,  inarr._byteoffset,  inarr._strides,
                   toutarr._data, toutarr._byteoffset, toutarr._strides)
            errorstatus = _numarray.CheckFPErrors()
            if errorstatus:
                numarray.handleError(errorstatus, sourcemsg)
        else:
            # slow case
            # And now a few words about overlap. It is used to allow the
            # accumulate C function to work across blocks. Overlap in effect
            # makes the first subarray of the next block the last subarray
            # of the previous block (or last array value for 1-d arrays) so
            # the accumulate can carry on. The computation of the block size
            # must take it into account as does computeOverDimensions in its
            # computation of the offsets for the blocks. This is only needed
            # where the accumulated dimension is larger than a block. It is
            # not needed at all for reduce since the output array contains
            # the net accumulation.
            overlap = 0
            
            # check special case where last dimension is bigger than blocksize
            maxitemsize = max(intype.bytes, outtype.bytes)
            niter = _bufferPool.buffersize/maxitemsize
            
            # This is a bit tricky. If the accumulated dimension is bigger
            # than the block size, it is necessary to overlap the block
            # computations by one element so that the last value of the
            # previous block can continue the count into the next block.
            if inarr._shape[-1] > niter:
                overlap = 1

            indexlevel, blockingparameters = \
                _getBlockingParameters(inarr._shape, niter, overlap)

            # The usual compute stuff
            input = _InputConverter(inarr, outtype=intype.name)
            output = _OutputConverter(toutarr, intype=outtype.name)
            preprocessing_outbuffers = [input.outbuffer]
            postprocessing_inbuffers = [output.inbuffer]
            operator = _Operator(ucfunc,
                 preprocessing_outbuffers, postprocessing_inbuffers,
                 stridefunction=outtype.bytes)

            functions = (input.compute, operator.compute, output.compute)
            _numarray.CheckFPErrors()
            _doOverDimensions(functions, toutarr._shape,
                              [], indexlevel, blockingparameters, overlap)
            errorstatus = _numarray.CheckFPErrors()
            if errorstatus:
                numarray.handleError(errorstatus, sourcemsg)

        if cumoperation=="reduce":
            toutarr._strides = toutarr._strides[:-1]
            if outarr is not None and outarr._type != outtype:
                _copyFromAndConvert(toutarr, outarr)
                return outarr
        return toutarr

    def accumulate(self, inarr, dim=0, outarr=None):
        """accumulate(self, inarr, dim=0, outarr=None) applies the binary
        operator ufunc 'self' at successive pairs of elements of 'inarr' along
        dimension 'dim', storing the result in 'outarr'.  If no 'outarr' is
        provided, the result of the accumulation will be returned.  The
        result of accumulating 'inarr' has the same shape as 'inarr'.
        """
        _inarr = numarray.swapaxes(inarr, -1, dim)
        _outarr0 = numarray.swapaxes(outarr, -1, dim)
        _outarr1 = self._cumulative("accumulate", _inarr, _outarr0)
        _outarr1.swapaxes(-1, dim) # swap axes to "fix" the result
        if outarr is None:
            return _outarr1

    def areduce(self, inarr, dim=0, outarr=None):
        """areduce(self, inarr, dim=0, outarr=None) applies the operator of
        ufunc 'self' to reduce the array 'inarr' along its 'dim' axis.
        If no 'outarr' is provided, the result of the reduction is returned.
        The result of reducing an N-dimensional array is an N-1 dimensional
        array.  The result of reducing a rank-1 array is a rank-0 array.

        see also 'reduce'.
        """
        _inarr = numarray.swapaxes(inarr, -1, dim)
        _outarr0 = numarray.swapaxes(outarr, -1, dim)
        _outarr1 = self._cumulative("reduce", _inarr, _outarr0)
        _outarr1.swapaxes(-1, dim) # swap axes to "fix" the result
        if outarr is None:
            return _outarr1

    def reduce(self, inarr, dim=0, outarr=None):
        """areduce(self, inarr, dim=0, outarr=None) applies the operator of
        ufunc 'self' to reduce the array 'inarr' along its 'dim' axis.
        If no 'outarr' is provided, the result of the reduction is returned.
        The result of reducing an N-dimensional array is an N-1 dimensional
        array.  The result of reducing a rank-1 array is a scalar.

        see also 'areduce'.
        """
        r = self.areduce(inarr, dim, outarr)
        if outarr is None:
            if r._shape == ():
                return r[0]
            else:
                return r

class _InputConverter:

    """This class handles alignment, byteswaping, copying and type conversions

    Stride here is a bit of a misnomer, it refers to one of
    functions (three currently) that deal with strides. These are:
    byteswap, align, copy operations. The conv functions are type conversion
    operations. They never deal with strides. At most one of each is
    necessary; neither is required. Strides and offsets are *byte*!

    """

    def __init__(self, inarr, outtype):
        self.stridefunction = None
        self.convfunction = None
        # Since the input array is iterated over dimensions, even
        # the output array, always contiguous, needs strides defined.
        self.outbytestrides =\
                _stridesFromShape(inarr._shape, inarr._itemsize)
        if inarr.isbyteswapped():
            if isinstance(inarr, numarray.ComplexArray):
                fname = "byteswap" + inarr._type.name
            else:
                fname = "byteswap"+`inarr._itemsize`+"bytes"
            self.stridefunction = _bytes.functionDict[fname]
            ## self.stridefunction = inarr._type._conv.byteswap
        elif not inarr.isaligned():
            fname = "align"+`inarr._itemsize`+"bytes"
            self.stridefunction = _bytes.functionDict.get(fname) or _bytes.functionDict["copyNbytes"]
        elif not inarr.iscontiguous():
            fname = "copy"+`inarr._itemsize`+"bytes"
            self.stridefunction = _bytes.functionDict.get(fname) or _bytes.functionDict["copyNbytes"]
        if outtype != inarr._type.name:
            self.convfunction = inarr._type._conv.astype[outtype]
        if self.stridefunction:
            self._itemsize = inarr._itemsize
            self.bytestrides = list(inarr._strides)
            buff1 = _bufferPool.getBuffer()
        if self.convfunction:
            buff2 = _bufferPool.getBuffer()
            if self.stridefunction:
                buffers = ((inarr, buff1), (buff1, buff2))
                outbuffer = buff2
            else:
                buffers = ((None, None), (inarr, buff2))
                outbuffer = buff2
        else:
            if self.stridefunction:
                buffers = ((inarr, buff1), (None, None))
                outbuffer = buff1
            else:
                buffers, outbuffer = ((None, None), (None, None)), inarr
        self.buffers = buffers
        self.outbuffer = outbuffer

    def compute(self, indices, shape):

        """ Perform the necessary transformations on the array.

        Results are in the last buffer used. Buffer management is
        done outside of this class. If no functions have been set
        nothing is done! offset and stride are applied to the
        first buffer (or last for reverse) encountered. The way
        it really works is that only one of the defined buffers
        (and at least one) is a real array. Only for that one
        will computeOffset generate a nonzero offset. Temporary
        buffers always have zero offset.
        """

        niter = ndarray.product(shape)
        if self.stridefunction:
            offset = self.buffers[0][0]._getByteOffset(indices)
            self.stridefunction(shape, 
                self.buffers[0][0]._data, offset, self.bytestrides[-len(shape):],
                self.buffers[0][1]._data, 0, self.outbytestrides[-len(shape):])
        if self.convfunction:
            offset = self.buffers[1][0]._getByteOffset(indices)
            self.convfunction(niter, 1, 1, ((self.buffers[1][0]._data, offset),
                 (self.buffers[1][1]._data, 0)))

class _OutputConverter:

    """This class handles alignment, byteswaping, copying and type

    conversions. Stride here is a bit of a misnomer, it refers to one of
    functions (three currently) that deal with strides. These are:
    byteswap, align, copy operations. The conv functions are type conversion
    operations. They never deal with strides. At most one of each is
    necessary; neither is required. Strides and offsets are *byte*!
    """
    def __init__(self, outarr, intype, inbuffer=None, forcestride=0):
        self.arr = outarr
        self.stridefunction = None
        self.convfunction = None
        # self._debug = (outarr, intype, inbuffer, forcestride)
        # Since the output array is iterated over dimensions, even
        # the input array, always contiguous, needs strides defined.
        # But first see if input buffer is a real array.
        if (inbuffer is not None) and inbuffer._strides:
            if not inbuffer.is_c_array():
                raise ValueError("input buffer is not a c_array")
            self.inbytestrides = inbuffer._strides
        else:
            self.inbytestrides =\
                _stridesFromShape(outarr._shape, outarr._itemsize)
        if outarr.isbyteswapped():
            if isinstance(outarr, numarray.ComplexArray):
                fname = "byteswap" + outarr._type.name
            else:
                fname = "byteswap"+`outarr._itemsize`+"bytes"
            self.stridefunction = _bytes.functionDict[fname]
        elif not outarr.isaligned():
            fname = "align"+`outarr._itemsize`+"bytes"
            self.stridefunction = _bytes.functionDict[fname]
        elif not outarr.iscontiguous() or forcestride:
            fname = "copy"+`outarr._itemsize`+"bytes"
            self.stridefunction = _bytes.functionDict[fname]
        if self.stridefunction:
            self.bytestrides = list(outarr._strides)
        if intype != outarr._type.name:
            self.convfunction = \
                          _numtypedict[intype]._conv.astype[outarr._type.name]
        # If inbuffer not specified, allocate one
        if (inbuffer is None) and (self.stridefunction or self.convfunction):
            inbuffer = _bufferPool.getBuffer()
        if self.convfunction:
            if self.stridefunction:
                buff = _bufferPool.getBuffer()
                buffers = ((inbuffer, buff), (buff, outarr))
            else:
                buffers = ((inbuffer, outarr), (None, None))
        else:
            if self.stridefunction:
                buffers = ((None, None),(inbuffer, outarr))
            else:
                if inbuffer is not None:
                    raise ValueError(
                     "Specified both input and output but no conversion implied")
                buffers, inbuffer = ((None, None), (None, None)), outarr

        self.buffers = buffers
        self.inbuffer = inbuffer


    def compute(self, indices, shape):

        """ perform the necessary transformations on the array.

        Results are in the last buffer used. Buffer management is
        done outside of this class. If no functions have been set
        nothing is done! offset and stride are applied to the
        last encountered. Offset is applied to the input buffer
        if is a real array.
        """

        niter = ndarray.product(shape)
        if self.convfunction:
            offset = self.buffers[0][1]._getByteOffset(indices)
            inoffset = self.buffers[0][0]._getByteOffset(indices)
            self.convfunction(niter, 1, 1, 
                              ((self.buffers[0][0]._data, inoffset),
                               (self.buffers[0][1]._data, offset)))
        if self.stridefunction:
            offset = self.buffers[1][1]._getByteOffset(indices)
            inoffset = self.buffers[1][0]._getByteOffset(indices)
            self.stridefunction(shape, 
                                self.buffers[1][0]._data, inoffset,
                                self.inbytestrides[-len(shape):],
                                self.buffers[1][1]._data, offset,
                                self.bytestrides[-len(shape):])

class _Operator:

    """performs the operation"""

    def __init__(self, cfunction, inputs, outputs, stridefunction=0):
        self.cfunction = cfunction
        self.inputs = inputs
        self.outputs = outputs
        self.stridefunction = stridefunction

    def compute(self, indices, shape):
        niter = ndarray.product(shape)
        args = []
        for item in self.inputs+self.outputs:
            if item._strides:
                offset = item._getByteOffset(indices)
            else:
                offset = 0
            args.append((item._data, offset))
            if self.stridefunction:
                if item._strides:
                    args.append(item._strides)
                else:
                    args.append(
                        _stridesFromShape(shape, self.stridefunction))
        if not self.stridefunction:
            self.cfunction(niter, len(self.inputs), len(self.outputs),
                           tuple(args))
        else:
            self.cfunction(shape, args[0][0], args[0][1], args[1],
                           args[2][0], args[2][1], args[3])

def _stridesFromShape(shape, bytestride):

    """Compute the strides from shape for a contiguous array, sort of"""
    if shape != ():
        ndim = len(shape)
        strides = [0]*ndim
        strides[-1] = bytestride
        for i in xrange(ndim-2, -1, -1):
            strides[i] = strides[i+1] * shape[i+1]
    else:
        strides = ()
    return strides

def _makeCUFuncDict(functionDict):
    """Organize C functions by function name and signature"""
    dict = {}
    for keystr in functionDict.keys():
        operator, optype, signature = eval(keystr)
        if not dict.has_key(operator):
            dict[operator] = {}
        if not dict[operator].has_key(optype):
            dict[operator][optype] = {}
        dict[operator][optype][signature[0]] = (signature[1],functionDict[keystr])
    return dict

def _isScalar(x):
    return numarray.PyNUMERIC_TYPES.has_key(type(x))

def _maxPopType(xs):
    """_maxPopType determines the maximum type of a sequence of
    arrays and scalars.  see numerictypes.genericCoercions.
    """
    if isinstance(xs[0], types.InstanceType) and \
       not isinstance(xs[0], numarray.NumArray):
        return None  # If they're not all NumArrays, quit now.

    maxT = None
    for x in xs:
        if isinstance(x, numarray.NumArray):
            t = x._type
        elif numarray.PyNUMERIC_TYPES.has_key(type(x)):
            t = type(x)
        else:
            raise TypeError( "Type of '%s' not supported" % repr(x) )
            
        if maxT is None:
            maxT = t
        else:
            maxT = numerictypes.genericCoercions[(maxT, t)]
    return maxT

CLIP = 0    # Peg indices > N-1 at N-1, indices < 0 at 0
WRAP = 1    # Index = Index % N

class _ChooseUFunc(_UFunc):

    """Class for building the "choose" ufunc.
    Class for ufuncs with 2 input (S, [~S...]) and 1 output (S) arguments"""

    def __init__(self, operator = "choose"):
        self.operator = operator

    def __call__(self, inarr1, inarr2, outarr=None, clipmode=CLIP):
        """The standard calling interface for UFuncs"""
        computation_mode, woutarr, cfunc, ufargs = \
                          self._setup(inarr1, inarr2, outarr, clipmode)

        result = self._doit(computation_mode, woutarr, cfunc, ufargs, 0)
        if outarr is None:
            return result
        else:
            return None

    def _setup(self, in1, in2, outarr=None, clipmode=CLIP):
        """Setup for Nary ufunc"""

        in1 = numarray.inputarray(in1, type=Long)
        in2 = list(in2)

        if outarr is None:
            convType = _maxPopType(in2)
            outarr = in1.new(convType)
        else:
            convType = outarr._type

        for i in range(len(in2)):
            if _isScalar(in2[i]):
                in2[i] = numarray.array([in2[i]], type=convType)
            elif ((type(in2[i]) == types.ListType) or
                  (type(in2[i]) == types.TupleType)):
                in2[i] = numarray.array(in2[i], type=convType)
            in2[i] = in1._broadcast(in2[i], raiseOnFail=1)

        if in1._shape != outarr._shape:
            raise ValueError("Supplied output array does not have appropriate shape")
        N = outarr._itemsize
        if N in [1,2,4,8,16]:
            ucfname = "choose" + `N` + "bytes"
            ucfunc = _bytes.functionDict[ucfname]
        else:
            ucfunc = _bytes.functionDict["chooseNbytes"]

        # pars = clipmode, population count, itemsize
        pars = _pseudoArrayFromScalars([clipmode, len(in2),N],
                                       numerictypes.Long)
        args = [pars, in1] + in2 + [outarr]
        fastargs = reduce(and_,
            [ _noConversionNeeded(x, convType) for x in in2 + [outarr]])
        fastargs = fastargs and _noConversionNeeded(in1, numerictypes.Long)
        if fastargs:
            # If fast case, return the arguments for an immediate call to the
            #  C function
            assert outarr._byteoffset == 0
            fastparms = tuple([ (x._data, x._byteoffset) for x in args])
            return ("fast", outarr, ucfunc,
                    (outarr.nelements(), len(args)-1, 1, fastparms))

        # slower, general case
        cpars         = _InputConverter(pars, numerictypes.Long)
        selector      = _InputConverter(in1, numerictypes.Long)
        population    = [_InputConverter(x, convType) for x in in2 ]

        inputs = (cpars, selector,) + tuple(population)
        outputs = (_OutputConverter(outarr=outarr, intype=convType),)
        maxitemsize = max(_numtypedict[convType].bytes,
                       numerictypes.Long.bytes)  # selector index.

        preprocessing_outbuffers = tuple([x.outbuffer for x in inputs])
        postprocessing_inbuffers = (outputs[0].inbuffer,)
        return ("slow", outarr, ucfunc, 
                (inputs, outputs, preprocessing_outbuffers,
                 postprocessing_inbuffers, maxitemsize))

choose = _ChooseUFunc()
choose.__doc__ ="""choose(selector, population) returns a new array shaped
like 'selector' with elements chosen from members of sequence 'population' by
the values of selector.  The shape of each member of 'population' must be
broadcastable to the shape of 'selector'.  The value of each member of
'selector' must satisfy:  0 <= value < len(population)."""

def _forwardBroadcast( indexArrays ):
    for i in range(len(indexArrays)-1):
        indexArrays[i], indexArrays[i+1] = \
                    indexArrays[i]._dualbroadcast(indexArrays[i+1], raiseOnFail=1)
    return indexArrays

def _nWayBroadcast( indexArrays ):
    # This is clear to me, if not super efficient.
    if len(indexArrays) > 1:
        indexArrays  = _forwardBroadcast( indexArrays )
        indexArrays.reverse()
        indexArrays  = _forwardBroadcast( indexArrays )
        indexArrays.reverse()
    return indexArrays

def _nonIteratedArray( array ):
    """_nonIteratedArray(array) returns a view of array which will not change
    offsets as a ufunc crosses buffer boundaries.
    """
    array = array.view() # make sure "_strides = None" doesn't leak out...
    array._strides = None
    return array

def _takeShape(scattered, indexArrays):
    nindexArrays = len(indexArrays)
    arrDims  = len(scattered._shape)

    if nindexArrays > arrDims:
        raise ValueError("Specified too many indices...")

    # Convert indices into suitable index arrays
    for i in range(len(indexArrays)):
        indexArrays[i] = numarray.inputarray(indexArrays[i], Long)

    indexArrays = _nWayBroadcast(indexArrays) # convert to arrays and broadcast

    # Figure out output array shape and basic blocksize N
    leftOver = arrDims - nindexArrays

    N = scattered._strides[ nindexArrays - 1 ] # size of inner block
    if nindexArrays < arrDims:
        nShape = scattered._shape[-leftOver:]
    else:
        nShape = ()

    impliedShape = indexArrays[0]._shape + nShape
    scatteredStrides = _nonIteratedArray(
        numarray.array( scattered._strides, type=Long ))
    scatteredShape   = _nonIteratedArray(
        numarray.array( scattered._shape,   type=Long ))

    return impliedShape, N, scatteredStrides, scatteredShape

class _TakeUFunc(_ChooseUFunc):
    """take(scattered, indexArrays, gathered=None)

    scattered:     input,        array from which elements are "gathered"

    indexArrays:   input,        tuple of index arrays or scalars

    gathered:      input/output  result array

    take plucks the elements of 'scattered' specified by the tuple of
    index arrays, 'indexArrays', and stores the result in the output array
    'gathered'.

    'indexArrays' may be partial, i.e.  incompletely specified,
    hence the shape of 'gathered' is derived from both the shape of
    'indexArrays' (primarily) and the shape of the trailing unspecified
    dimensions of 'scattered'.
    """
    def _setup(self, scattered, indexArrays, gathered=None, clipmode=CLIP):
        """Setup for Nary ufunc
        """
        scattered = numarray.inputarray(scattered)
        indexArrays = list(indexArrays)
        gatheredWasSpecified = (gathered is not None)
        if gatheredWasSpecified:
            ctype = gathered._type
        else:
            ctype = scattered._type

        # *All* of the scattered array must be simultaneously available.
        #    So blocked conversions won't work.
        if not _noConversionNeeded(scattered, ctype):
            scattered = scattered.astype(ctype)

        impliedShape, N, scatteredStrides, scatteredShape = \
                  _takeShape(scattered, indexArrays)

        if gatheredWasSpecified:
            if gathered._shape != impliedShape:
                raise ValueError("Inconsistent array shapes...")
        else:
            gathered = scattered.__class__(shape=impliedShape,
                                            type=scattered._type)
        # scattered and gathered should now be identically typed.

        # Since N can be arbitrarily large depending on the trailing
        # dimensions of scattered, may have to make gathered contiguous
        # to ensure a buffer large enough to store at least N.
        if (N > _bufferPool.buffersize
            and not _noConversionNeeded(gathered, scattered._type)):
            if gatheredWasSpecified:
                raise ValueError("Take can't take it anymore...")
            else:
                gathered = gathered.astype(scattered._type)

        niter = (gathered.nelements()*gathered._itemsize)/N
        ucfunc = _bytes.functionDict[self.operator + "Nbytes"]

        pars = _pseudoArrayFromScalars([clipmode, N], type=Long)

        fastargs = reduce(and_,
             [_noConversionNeeded(x, Long) for x in indexArrays ])
        fastargs = fastargs and _noConversionNeeded(gathered, scattered._type)
        if fastargs:
            args = [pars, scattered, scatteredStrides,
                    scatteredShape] + indexArrays + [gathered]
            fastparms = tuple([ (x._data, x._byteoffset) for x in args])
            return ("fast", gathered, ucfunc,
                    (niter, len(args)-1, 1, fastparms))

        scattered = _nonIteratedArray(scattered)

        # slower, general case
        # Note: both scattered and gathered converters are "null" converters.
        cPars               = _InputConverter(pars, Long)
        cScattered          = _InputConverter(scattered, gathered._type)
        cScatteredStrides   = _InputConverter(scatteredStrides, Long)
        cScatteredShape     = _InputConverter(scatteredShape, Long)
        cIndexArrays        = [_InputConverter(x, Long) \
                               for x in indexArrays ]
        inputs              = (cPars, cScattered, cScatteredStrides,
                               cScatteredShape) + tuple(cIndexArrays)
        outputs = (_OutputConverter(outarr=gathered, intype=gathered._type),)
        maxitemsize = max(N, Long.bytes)  # selector index.
        preprocessing_outbuffers = tuple([x.outbuffer for x in inputs])
        postprocessing_inbuffers = (outputs[0].inbuffer,)

        return ("slow", gathered, ucfunc,
                (inputs, outputs, preprocessing_outbuffers,
                 postprocessing_inbuffers, maxitemsize))

_take = _TakeUFunc("take")

def take(array, *indices, **keywords):
    """take(array, indices..., clipmode=(CLIP|WRAP), axis=0) picks elements
    of 'array' specified by index arrays 'indices'.

    parameters which must be specified by keyword:

    clipmode=CLIP   if clipmode is CLIP, out of range indices are clipped
                        at [0, shape[i]).

                    if clipmode is WRAP, out of range indices are wrapped
                        around from 0 to shape[i] (<0) or from shape[i] to
                        0 (>= shape[i])

    axis=0          selects the axis along which the take should be performed.

    """
    try:
        axis = keywords["axis"]
        del keywords["axis"]
    except KeyError:
        axis = 0
    if axis == 0:
        return _take(array, indices, **keywords)
    else:
        return numarray.swapaxes(_take(
            numarray.swapaxes(array, 0, axis), indices, **keywords), 0, axis)

class _PutUFunc(_TakeUFunc):
    """put(scattered, indexArrays, gathered) is the inverse function of "take",
    and scatters the elements of array 'gathered' across array 'scattered' as
    specified by tuple of index arrays, 'indexArrays'.

    scattered: input/output           Array where "stuff" gets put
    indexArrays: input                Arrays of indexArrays of scattered where
                                         corresponding pieces of gathered go.
    gathered: input                  Source array

    gathered can be an array which is broadcastable to the array shape
    implied by scattered and indexArrays  -or-  gathered can be a scalar.
    """
    def _setup(self, scattered, indexArrays, gathered, clipmode=CLIP):
        """Setup for Nary ufunc
        """
        scattered = numarray.inputarray(scattered)
        indexArrays = list(indexArrays)

        impliedShape, N, scatteredStrides, scatteredShape  = \
                  _takeShape(scattered, indexArrays)

        pars = _pseudoArrayFromScalars([clipmode, N], type=Long)

        gathered = numarray.inputarray(gathered, type=scattered._type)
        gathered = ndarray._broadcast(gathered, impliedShape, raiseOnFail = 1)

        if not gathered.iscontiguous():
            gathered = gathered.copy()

        # gathered and scattered should now be identically typed.
        # gathered need not be contiguous since it is "bufferable".
        # Since N can be arbitrarily large depending on the trailing
        # dimensions of scattered, may have to pre-process gathered.
        if N > _bufferPool.buffersize and \
           not _noConversionNeeded(gathered, scattered._type):
            gathered = gathered.astype( scattered._type )

        scattered = _nonIteratedArray(scattered)

        niter = (gathered.nelements()*gathered._itemsize)/N
        ucfunc = _bytes.functionDict[self.operator + "Nbytes"]

        fastargs = reduce(and_,
             [ _noConversionNeeded(x, Long) for x in indexArrays ])
        fastargs = fastargs and _noConversionNeeded(gathered, scattered._type)
        if fastargs:
            args = [pars, gathered, scatteredStrides,
                    scatteredShape] + indexArrays + [scattered]
            fastparms = tuple([ (x._data, x._byteoffset) for x in args])
            return ("fast", scattered, ucfunc, (niter, len(args)-1, 1,
                                                fastparms))
        # slower, general case

        # Since scattered is not bufferable because of the random nature of
        # indexing, scattered is processed as a conversion-less input.
        # Note: scattered is input/output.
        cPars               = _InputConverter(pars, Long)
        cScattered          = _InputConverter(scattered, scattered._type)
        cScatteredStrides   = _InputConverter(scatteredStrides, Long)
        cScatteredShape     = _InputConverter(scatteredShape, Long)
        cIndexArrays        = [_InputConverter(x, Long) \
                               for x in indexArrays ]
        cGathered          = _InputConverter(gathered, scattered._type)
        inputs = (cPars, cGathered, cScatteredStrides, cScatteredShape) + \
                 tuple(cIndexArrays)
        outputs = (cScattered,)
        maxitemsize = max(N, Long.bytes)  # selector index.
        preprocessing_outbuffers = tuple([x.outbuffer for x in inputs])
        postprocessing_inbuffers = (cScattered.outbuffer,)

        # Note: gathered is specified as "output" to drive shape agreement
        # between gathered and all indexArrays.  This works since (a) inputs
        # are writeable and (b) "put" doesn't return any result.
        return ("slow", gathered, ucfunc, 
                (inputs, outputs, preprocessing_outbuffers,
                 postprocessing_inbuffers, maxitemsize))

    def _doit(self, computation_mode, outarr, cfunc, ufargs, scalar):
        _UFunc._doit(self, computation_mode, outarr, cfunc, ufargs, scalar)

_put = _PutUFunc("put")

def put(array, *indices, **keywords):
    """put(array, indices..., values, clipmode=(CLIP|WRAP), axis=0) stores 'values'
    into 'array' at 'indices...'.

    parameters which must be specified by keyword:

    clipmode=CLIP   if clipmode is CLIP, out of range indices are clipped
                        at [0, shape[i]).

                    if clipmode is WRAP, out of range indices are wrapped
                        around from 0 to shape[i] (<0) or from shape[i] to
                        0 (>= shape[i])

    axis=0          selects the axis along which the take should be performed.

    """
    try:
        axis = keywords["axis"]
        del keywords["axis"]
    except KeyError:
        axis = 0
    if axis == 0:
        return _put(array, indices[:-1], indices[-1], **keywords)
    else:
        return numarray.swapaxes( _put( numarray.swapaxes(array, 0, axis),
                               indices[:-1], indices[-1], **keywords),
                         0, axis)

class _NonzeroUFunc:
    """nonzero(array)

    array: input,        array scanned for non-zero elements

    Nonzero scans "array" for nonzero elements, and returns a tuple
    of coordinate arrays corresponding to the nonzero elements.

    """
    def __call__(self, inarr1):

        """The standard calling interface for UFuncs"""
        computation_mode, outarr, cfunc, ufargs = self._setup(inarr1)
        apply(cfunc, ufargs)
        return tuple(outarr)

    def _setup(self, inarr1):
        """Setup for Nonzero
        """
        nz = not_equal(inarr1, 0)
        nonzeroCount = add.reduce(ndarray.ravel(nz).astype(Long))
        outarr = [ numarray.NumArray(shape=(nonzeroCount,), type=Long)
                   for i in range(nz.getrank()) ]
        sstrides = numarray.array(nz._strides, type=Long)
        sstrides /= nz._bytestride

        niter = nz.nelements()
        ucfunc = _sort.functionDict[repr(("Bool", "nonzeroCoords"))]

        args = [ nz, sstrides] + outarr
        fastparms = tuple([ (x._data, x._byteoffset) for x in args])
        return ("fast", outarr, ucfunc, (niter, 2, len(args)-2, fastparms))


def _sort1(inarr1):
    """1D in-place Sort"""
    conversionNeeded = not _noConversionNeeded(inarr1, inarr1._type)
    contigarr1 = ((conversionNeeded and numarray.array(inarr1)) or inarr1)
    cfunc = _sort.functionDict[repr((contigarr1._type.name, 'sort'))]
    cfunc(contigarr1.nelements(), 0, 1,
          ((contigarr1._data, contigarr1._byteoffset),))
    if conversionNeeded:
        inarr1._copyFrom(contigarr1)
 
def searchsorted(bins, values):
    """searchsort(bins, values) returns the array of greatest indices 'i'
    such that each values[i] <= bins[i].
    """
    bins = numarray.inputarray(bins)
    values = numarray.inputarray(values)
    if len(bins._shape) > 1 or len(values._shape) > 1:
        raise ValueError("Multi-dimensional searchsort not supported.")
    outarr = numarray.NumArray(shape=len(values), type=Long)
    if not _noConversionNeeded(bins, bins._type):
        bins = bins.copy()
    if not _noConversionNeeded(values, bins._type):
        values = values.astype(bins._type)
    cfunc = _sort.functionDict[repr((bins._type.name, 'searchsorted'))]
    nbins = _pseudoArrayFromScalars([len(bins)], Long)
    args = [(x._data, x._byteoffset) for x in [nbins, bins, values, outarr]]
    cfunc(values.nelements(), 3, 1, tuple(args))
    return outarr

def _argsort1(inarr1, witness=None):
    """1D ArgSort"""
    niter = inarr1.nelements()
    inarr1 = numarray.array(inarr1) # make a discardable copy
    if witness is None:
        witness =numarray.arange(niter,type=Long)
    elif not _noConversionNeeded(witness, Long):
        raise ValueError("No support for converting the witness array")
    cfunc = _sort.functionDict[repr((inarr1._type.name,'asort'))]
    cfunc(niter, 0, 2,
          tuple([ (x._data, x._byteoffset) for x in [inarr1, witness]]))
    return witness

# =====================================================================
# import these last so module dependencies don't cause problems
import numerictypes
import numarray

Long = numerictypes.Long

# module-level variables
_numtypedict = numerictypes.typeDict

# Short term solution -- identity needs to be defined in _ufunc module XXX
_identities = {"add":0, "subtract":0, "multiply":1, "divide":1}
_cufuncs = _makeCUFuncDict(_ufunc.functionDict)

_UFuncs = {}
for operator in _cufuncs.keys():
    _UFuncs[operator] = ufuncFactory(operator, _cufuncs[operator],
                            identity=_identities.get(operator))
globals().update(_UFuncs)
del operator, _UFuncs

_bufferPool = _BufferPool()
nonzero = _NonzeroUFunc()

def _fbroadcast(f, N, args, shape):
    """_fbroadcast(f, N, args, shape) calls 'f' for each of the
    'N'-dimensional inner subarrays of 'args'.  Each subarray has
    .shape == 'shape'[-N:].  There are a total of product(shape[:-N])
    calls to 'f'.
    """
    if len(shape) == N:
        apply(f, tuple(args))
    else:
        for i in range(shape[0]):
            _fbroadcast(f, N, [x[i] for x in args], shape[1:])
        
def _sortN(a):
    """_sortN implements N-D in-place sort in terms of 1D sort"""
    _fbroadcast(_sort1, 1, (a,), a._shape)

def _argsortN(a, w):
    """_argsortN implements N-D in-place argsort in terms of 1D argsort"""
    _fbroadcast(_argsort1, 1, (a,w), a._shape)

def _chktype(a):
    if isinstance(a._type, numerictypes.IntegralType):
        artype = numerictypes.Long
    elif isinstance(a._type, numerictypes.FloatingType):
        artype = numerictypes.Float
    elif isinstance(a._type, numerictypes.ComplexType):
        artype = numerictypes.Complex
    else:
        raise ValueError("Only integer, float, and complex matrixmultiply supported.")
    return artype

def _matrixtype(a,b):
    artype, brtype = _chktype(a), _chktype(b)
    if artype != brtype:
        return numerictypes.Float64
    else:
        return artype

def innerproduct(a,b):
    """innerproduct(a,b) computes dot products of the last dimensions of
    MxL matrix 'a' and NxL matrix 'b' resulting in an MxN matrix.

    innerproduct coerces its inputs to the maximum type of the inputs'
    typeclass, either Long, Float, or Complex
    """
    a = numarray.inputarray(a)
    b = numarray.inputarray(b)
    if a._shape[-1] != b._shape[-1]:
        raise ValueError("incompatible array dimensions.")

    rtype = _matrixtype(a,b)

    if not _noConversionNeeded(a, rtype):
        a = a.astype(rtype)
    if not _noConversionNeeded(b, rtype):
        b = b.astype(rtype)

    adots, bdots = a._shape[:-1], b._shape[:-1]
    r = a.__class__(shape=adots+bdots, type=rtype)

    maxijk = _pseudoArrayFromScalars([ndarray.product(adots),
                                    ndarray.product(bdots),
                                    a._shape[-1]],
                                     Long)

    _numarray.CheckFPErrors()

    cfunc = _sort.functionDict[repr((rtype.name, "mxmultiply"))]
    cfunc(1, 3, 1, tuple([(x._data, x._byteoffset) for x in \
                          [ a, b, maxijk, r]]))

    errorstatus = _numarray.CheckFPErrors()
    if errorstatus:
        numarray.handleError(errorstatus, " in innerproduct")

    if len(adots+bdots):
        return r
    else:
        return r[0]

