#include <stdio.h>
#include <math.h>
#include <signal.h>
#include <ctype.h>

#include <Python.h>

#include "arrayobject.h"

static PyObject *_Error;

#define A_GET1(a, t, i)  (*((t *) ((a)->data+((i)*(a)->strides[0]))))
#define A_SET1(a, t, i, v)  (A_GET1(a, t, i) = (v))

#define A_GET2(a, t, i, j)  (*((t *) ((a)->data+((i)*(a)->strides[0] +        \
                                                 (j)*(a)->strides[1]))))
#define A_SET2(a, t, i, j, v)  (A_GET2(a, t, i, j) = (v))


static void 
Convolve1d(PyArrayObject *kernel, PyArrayObject *data, 
	   PyArrayObject *convolved)
{
	int xc, xk;
	int ksizex = kernel->dimensions[0];
	int halfk = ksizex / 2;
	int dsizex = data->dimensions[0];

	for(xc=0; xc<halfk; xc++)
		A_SET1(convolved, Float64, xc, A_GET1(data, Float64, xc));
		     
	for(xc=dsizex-halfk; xc<dsizex; xc++)
		A_SET1(convolved, Float64, xc, A_GET1(data, Float64, xc));

	for(xc=halfk; xc<dsizex-halfk; xc++) {
		Float64 temp = 0;
		for (xk=0; xk<ksizex; xk++) {
			int i = xc - halfk + xk;
			Float64 k, d;
			k = A_GET1(kernel, Float64, xk);
			d = A_GET1(data, Float64, i);
			temp += k * d;
		}
		A_SET1(convolved, Float64, xc, temp);
	}
}

static void 
Convolve2d(PyArrayObject *kernel, PyArrayObject *data, 
	   PyArrayObject *convolved)
{
	int ki, kj, di, dj;
	int krows = kernel->dimensions[0], kcols = kernel->dimensions[1];
	int drows = data->dimensions[0], dcols = data->dimensions[1];
	int halfkrows = krows/2;
	int halfkcols = kcols/2;

	/* Copy the data in the half kernel "frame" straight through. */
	for(di=0; di<halfkrows; di++) {
		for(dj=0; dj<dcols; dj++)
			A_SET2(convolved, Float64, di, dj,
				A_GET2(data, Float64, di, dj));
	}
	for(di=drows-halfkrows; di<drows; di++) {
		for(dj=0; dj<dcols; dj++)
			A_SET2(convolved, Float64, di, dj,
				A_GET2(data, Float64, di, dj));
	}
	for(di=halfkrows; di<drows-halfkrows; di++) {
		for(dj=0; dj<halfkcols; dj++)
			A_SET2(convolved, Float64, di, dj,
				A_GET2(data, Float64, di, dj));
	}
	for(di=halfkrows; di<drows-halfkrows; di++) {
		for(dj=dcols-halfkcols; dj<dcols; dj++)
			A_SET2(convolved, Float64, di, dj,
				A_GET2(data, Float64, di, dj));
	}

	for(di=halfkrows; di<drows-halfkrows; di++) {
		for(dj=halfkcols; dj<dcols-halfkcols; dj++) {
			Float64 temp = 0;
			for(ki=0; ki<krows; ki++) {
				int pi = di + ki - halfkrows;
				for(kj=0; kj<kcols; kj++) {
					int pj = dj + kj - halfkcols;
					temp += 
					     A_GET2(data, Float64, pi, pj) *
					     A_GET2(kernel, Float64, ki, kj);
				}
			}
			A_SET2(convolved, Float64, di, dj, temp);
		}
	}
}

static PyObject *
Py_Convolve1d(PyObject *obj, PyObject *args)
{
	PyObject   *kernel, *data, *convolved=Py_None;
	PyArrayObject *kem, *dem, *cem;

	if (!PyArg_ParseTuple(args, "OO|O", &kernel, &data, &convolved))
		return PyErr_Format(_Error, 
				    "Convolve2d: Invalid parameters.");

	kem = (PyArrayObject *) PyArray_ContiguousFromObject(kernel, 
							     tFloat64, 1, 1);
	dem = (PyArrayObject *) PyArray_ContiguousFromObject(data, 
							     tFloat64, 1, 1);
	if (!kem || !dem) return NULL;

	if (convolved == Py_None)
		cem = (PyArrayObject *) PyArray_FromDims(
			dem->nd, dem->dimensions, tFloat64);
	else
		cem = (PyArrayObject *) PyArray_ContiguousFromObject(
			convolved, tFloat64, 1, 1);
	if (!cem) return NULL;

	Convolve1d(kem, dem, cem);

	Py_DECREF(kem);
	Py_DECREF(dem);
	if (convolved != Py_None) {
		Py_DECREF(cem);
		Py_INCREF(Py_None);
		return Py_None;
	} else
		return PyArray_Return(cem);
}

static PyObject *
Py_Convolve2d(PyObject *obj, PyObject *args)
{
	PyObject   *kernel, *data, *convolved=Py_None;
	PyArrayObject *kem, *dem, *cem;

	if (!PyArg_ParseTuple(args, "OO|O", &kernel, &data, &convolved))
		return PyErr_Format(_Error, 
				    "Convolve2d: Invalid parameters.");

	kem = (PyArrayObject *) PyArray_ContiguousFromObject(
		kernel, tFloat64, 2, 2);
	dem = (PyArrayObject *) PyArray_ContiguousFromObject(
		data, tFloat64, 2, 2);

	if (!kem || !dem) return NULL;

	if (convolved == Py_None)
		cem = (PyArrayObject *) PyArray_FromDims(
			dem->nd, dem->dimensions, tFloat64);
	else
		cem = (PyArrayObject *) PyArray_ContiguousFromObject(
			convolved, tFloat64, 2, 2);
	if (!cem) return NULL;

	Convolve2d(kem, dem, cem);

	Py_DECREF(kem);
	Py_DECREF(dem);
	if (convolved != Py_None) {
		Py_DECREF(cem);
		Py_INCREF(Py_None);
		return Py_None;
	} else
		return PyArray_Return(cem);
}

static PyMethodDef _convolveMethods[] = {
	{"Convolve1d", Py_Convolve1d, METH_VARARGS, 
	 "Convolve1d(kernel, data, output=None) convolves 1d array 'kernel' " \
	 "with 1d array 'data' either returning the result or silently storing " \
	 "it in 'output'."},
	{"Convolve2d", Py_Convolve2d, METH_VARARGS,
	 "Convolve2d(kernel, data, output=None) convolves 2d array 'kernel' " \
	 "with 2d array 'data' either returning the result or silently storing " \
	 "it in 'output'."},
	{NULL, NULL} /* Sentinel */
};

/* platform independent*/
#ifdef MS_WIN32
__declspec(dllexport)
#endif

void initnumpy_compat(void) {
	PyObject *m, *d;
	m = Py_InitModule("numpy_compat", _convolveMethods);
	d = PyModule_GetDict(m);
	_Error = PyErr_NewException("_numpy_compat.error", NULL, NULL);
	PyDict_SetItemString(d, "error", _Error);
	import_array();
}



/*
 * Local Variables:
 * mode: C
 * c-file-style: "python"
 * End:
 */
